@@ -30,6 +30,8 @@ def logWrapper(fn):
         logMessage(u'entering %s' % fn.__name__)
         logMessage(u'args: %s' % unicode(args))
         logMessage(u'kwargs: %s' % unicode(kwargs))
-        fn(*args, **kwargs)
-        logMessage(u'exiting %s' % fn.__name__)
+        try:
+            return fn(*args, **kwargs)
+        finally:
+            logMessage(u'exiting %s' % fn.__name__)
     return newFn
