@@ -7,15 +7,19 @@
 # 
 # A copy of the license has been included in the COPYING file.
 
-import os, sys, re, pexpect, random
+import os, sys, re, random
 import string, locale, struct, fcntl, termios, signal
 
+from pexpect import (
+  spawn,
+  TIMEOUT,
+  EOF,
+)
+
 from sclapp import shinterp
 from sclapp.exceptions import Error
 from sclapp.util import DelegateWrapper, CallbackMapping
-
-TIMEOUT = pexpect.TIMEOUT
-EOF = pexpect.EOF
+from sclapp.error_output import printError
 
 RAND_PROMPT_SIZE = 32
 
@@ -36,18 +40,6 @@ class CommandFinished(Error):
 class ExitInteract(Error):
     pass
 
-class Pexpect(DelegateWrapper):
-    def __init__(self, cmd, **kwargs):
-        self.wrapped_obj = pexpect.spawn(cmd, **kwargs)
-
-        # Monkey patch inappropriate __del__ application in pexpect.  We expect
-        # callers to call our close method.  If they don't, it's their own damn
-        # fault.
-        if hasattr(pexpect.spawn, '__del__'):
-            del pexpect.spawn.__del__
-
-        super(Pexpect, self).__init__()
-
 def lengthOfLongestLine(s):
     length = 0
     for line in s.split('\n'):
@@ -56,7 +48,7 @@ def lengthOfLongestLine(s):
             length = len_line
     return length
 
-class Shell(Pexpect):
+class Shell(spawn):
     shell = None
     _prompt = None
 
@@ -122,10 +114,22 @@ class Shell(Pexpect):
         self.env['PS3'] = ''
         self.env['PS4'] = ''
 
+        if 'TERM' in os.environ:
+            self.env['TERM'] = os.environ['TERM']
+
         self.always_group_cmd = always_group_cmd
 
         self.setupShell()
 
+    def __del__(self):
+        # Make an effort to clean up here just in case our caller neglected to
+        # do so.  If we don't, bad things (like not closing the pty) can
+        # happen.
+        try:
+            self.exit()
+        except Exception, e:
+            printError(e)
+
     def setupShell(self):
         self.execute('set +m')
 
@@ -165,13 +169,13 @@ class Shell(Pexpect):
             self.timeout = new_timeout
         return old_timeout
 
-    def _execute(self, cmd, *values, **kwargs):
+    def _interpolate_command(self, cmd, *values, **kwargs):
+        return self.interpolate(cmd.strip(), *values)
+
+    def _adjust_command(self, interpolated_cmd, *values, **kwargs):
         force = kwargs.get('force', False)
 
-        adjusted_cmd = cmd.strip()
-        adjusted_cmd = self.interpolate(adjusted_cmd, *values)
-
-        interpolated_command = adjusted_cmd
+        adjusted_cmd = interpolated_cmd
 
         if self.always_group_cmd:
             adjusted_cmd = '{ %s\n}' % adjusted_cmd
@@ -186,8 +190,16 @@ class Shell(Pexpect):
               'Use force = True if you really want this.'
             )
 
+        return adjusted_cmd
+
+    def _execute(self, cmd, *values, **kwargs):
+        interpolated_cmd = self._interpolate_command(
+          cmd, *values, **kwargs)
+        adjusted_cmd = self._adjust_command(
+          interpolated_cmd, *values, **kwargs)
+        
         if self.trace:
-            trace_cmd = interpolated_command
+            trace_cmd = interpolated_cmd
             if not isinstance(trace_cmd, unicode):
                 trace_cmd = self.decode(trace_cmd)
             print trace_cmd
@@ -197,7 +209,7 @@ class Shell(Pexpect):
             bytes_sent = self.sendline(adjusted_cmd)
         finally:
             if bytes_sent > len(adjusted_cmd):
-                self._last_cmd = interpolated_command
+                self._last_cmd = interpolated_cmd
 
     def _getLastOutput(self):
         self._pushTimeout(MAX_DELAY)
@@ -494,16 +506,24 @@ class Shell(Pexpect):
 
     def interact(self, fitted = False):
         if fitted:
-            old_handler = signal.signal(signal.SIGWINCH, self._sigwinch_handler)
-        try:
-            if fitted:
-                self.fit_window()
-            return self._interact()
-        finally:
-            if fitted:
-                signal.signal(signal.SIGWINCH, old_handler)
+            return self._interact_fitted()
+        return self._interact()
 
-    def _interact(self):
+    def interact_with(self, *args, **kwargs):
+        fitted = kwargs.pop('fitted', False)
+        if fitted:
+            return self._interact_fitted(*args, **kwargs)
+        return self._interact(*args, **kwargs)
+
+    def _interact_fitted(self, *args, **kwargs):
+        old_handler = signal.signal(signal.SIGWINCH, self._sigwinch_handler)
+        try:
+            self.fit_window()
+            return self._interact(*args, **kwargs)
+        finally:
+            signal.signal(signal.SIGWINCH, old_handler)
+
+    def _interact(self, cmd = None, *values, **kwargs):
         old_prompt = self.prompt
 
         self.execute('sh')
@@ -520,13 +540,22 @@ class Shell(Pexpect):
         self.sendline('PS4="%s"; export PS4' % self.interactive_ps4)
         self.expect(self.prompt_sep)
 
-        # Trigger first prompt to be displayed:
-        self.sendline()
+        if cmd is not None:
+            cmd = u'%s; exit $??' % cmd
+            interpolated_cmd = self._interpolate_command(
+              cmd, *values, **kwargs)
+            adjusted_cmd = self._adjust_command(
+              interpolated_cmd, *values, **kwargs)
+            self.sendline(adjusted_cmd)
+            self._last_cmd = interpolated_cmd
+        else:
+            # Trigger first prompt to be displayed:
+            self.sendline()
 
         self.setecho(True)
         try:
             try:
-                self.wrapped_obj.interact(
+                super(Shell, self).interact(
                   escape_character = chr(0),
                   output_filter = self._interact_output_filter,
                 )
@@ -534,10 +563,15 @@ class Shell(Pexpect):
                 pass
         finally:
             print ''
-            print 'Exiting interactive shell...'
             self.setecho(False)
         self._prompt = old_prompt
-        self.expect(self.prompt_sep)
+        try:
+            self.expect(self.prompt_sep, timeout = 2)
+        except TIMEOUT:
+            self.sendline()
+            self.expect(self.prompt_sep, timeout = 2)
+
+        return self._checkLastStatus(None)
 
 SORRY_TRY_AGAIN = 'Sorry, try again.'
 PASSWORD_PROMPT = 'Password:'
